(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('d3-selection'), require('d3-transition')) :
  typeof define === 'function' && define.amd ? define(['exports', 'd3-selection', 'd3-transition'], factory) :
  (factory((global.d3 = {}),global.d3,global.d3));
}(this, (function (exports,d3Selection,d3Transition) { 'use strict';

  function empty() {}
  function identity($) { return $; }

  function gup() {
    var pre = null
      , exit = null
      , enter = null
      , post = null
      , select = null
    ;
    function gup(data) {
      var more = [], len = arguments.length - 1;
      while ( len-- > 0 ) more[ len ] = arguments[ len + 1 ];

      if (data instanceof d3Selection.selection || data instanceof d3Transition.transition) {
        return gup.apply(this, more)(data);
      }
      function _gup() {
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

        var context = args.shift();
        var shouldTransition = !!context.selection;
        var selection = shouldTransition ? context.selection() : context;
        if (select && select != identity) {
          selection = select.call.apply(select, [ this, selection ].concat( args ));
        }

        var $pre = selection.data(data, more[0]);
        if (pre && pre != empty) {
          var preT = $pre;
          if (shouldTransition && $pre.transition) {
            preT = $pre.transition(context);
          }
          preT.call.apply(preT, [ pre ].concat( args ));
        }

        if (exit && exit != empty) {
          var $exit = $pre.exit();
          if (shouldTransition && $exit.transition) {
            $exit = $exit.transition(context);
          }
          $exit.call.apply($exit, [ exit ].concat( args ));
        }

        var $enter = $pre.enter();
        if (enter && enter != identity) {
          $enter = enter.call.apply(enter, [ this, $enter ].concat( args ));
        }

        var $post = $enter.merge($pre);
        if (shouldTransition && $post.transition) {
          $post = $post.transition(context);
        }
        if (post && post != empty) {
          $post.call.apply($post, [ post ].concat( args ));
        }

        return $post;
      }

      _gup.data = function() {
        var assign;

        var _ = [], len = arguments.length;
        while ( len-- ) _[ len ] = arguments[ len ];
        return arguments.length ? ((assign = _, data = assign[0], more = assign.slice(1), assign), this) : [data ].concat( more);
      };

      return _gup;
    }

    gup.select = function(_) {
      return arguments.length ? (select = _, this) : (select || identity);
    };

    gup.pre = function(_) {
      return arguments.length ? (pre = _, this) : (pre || empty);
    };

    gup.exit = function(_) {
      return arguments.length ? (exit = _, this) : (exit || empty);
    };

    gup.enter = function(_) {
      return arguments.length ? (enter = _, this) : (enter || identity);
    };

    gup.post = function(_) {
      return arguments.length ? (post = _, this) : (post || empty);
    };

    gup.update = function() {
      var assign, assign$1, assign$2, assign$3;

      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];
      switch (arguments.length) {
        case 0: return [
          pre || empty,
          exit || empty,
          enter || identity,
          post || empty
        ];
        case 1:
          (assign = args, pre = assign[0]);
          break;
        case 2:
          (assign$1 = args, pre = assign$1[0], exit = assign$1[1]);
          break;
        case 3:
          (assign$2 = args, pre = assign$2[0], exit = assign$2[1], enter = assign$2[2]);
          break;
        default:
          (assign$3 = args, pre = assign$3[0], exit = assign$3[1], enter = assign$3[2], post = assign$3[3]);
      }
      return this;
    };

    return gup;
  }

  var nonEnumerableProps = /^(valueOf|isPrototypeOf|to(Locale)?String|propertyIsEnumerable|hasOwnProperty|pre|exit|enter|post|update|select)$/;

  function comp(fns, name) {
    var forward = name === "enter" || name === "select";
    var f = function() {
      var this$1 = this;
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      for (var i = 0, list = fns; i < list.length; i += 1) {
        var f = list[i];

        if (name in f && f[name] && (f = f[name]()) && f !== empty && f !== identity) {
          var result = f.apply(this$1, args);
          if (forward) { args[0] = result; }
        }
      }
      if (forward) { return result; }
    };
    return f;
  }

  function compose() {
    var fns = [], len = arguments.length;
    while ( len-- ) fns[ len ] = arguments[ len ];

    var f = gup();

    for (var i = 0, list = fns; i < list.length; i += 1) {
      var source = list[i];

      for (var k in source) {
        if (!nonEnumerableProps.test(k)) {
          f[k] = source[k];
        }
      }
    }

    fns = fns.reverse();
    f.select(comp(fns, "select"));
    f.pre(comp(fns, "pre"));
    f.exit(comp(fns, "exit"));
    f.enter(comp(fns, "enter"));
    f.post(comp(fns, "post"));

    return f;
  }

  exports.gup = gup;
  exports.gupEmpty = empty;
  exports.gupIdentity = identity;
  exports.gupCompose = compose;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
